from django.conf.urls import patterns, url
from . import views, MAGIC_ID


_MODULE_MAGIC_ID_ = MAGIC_ID

urlpatterns = patterns('',
                       url(r'^login/$', views.login, name="sso_login"),
                       url(r'^reqeusttoken/$', views.api_requesttoken, name="sso_requesttoken"),
                       url(r'^authtoken/$', views.api_authtoken, name="sso_authtoken"))



import urlparse
import json
from functools import partial

from django.http import HttpResponse
from djapiauth.auth import api_auth
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.contrib.auth.forms import AuthenticationForm
try:
    from django.contrib.sites.shortcuts import get_current_site
except:
    from django.contrib.sites.models import get_current_site
from django.contrib.auth.models import AnonymousUser

try:
    from django.http import JsonResponse
except ImportError:
    from .utility import JsonResponse

from sso import gen_auth_token, RequestToken, SSO_REQUEST_TOKEN_TTL
from .models import SSO
from utility import jsonify, append_tokens

JsonResponse403 = partial(JsonResponse, status=403)


@api_auth
@jsonify
def api_requesttoken(request):
    return {"request_token": RequestToken.generate()}


@api_auth
def api_authtoken(request):
    _request_token = request.GET.get("request_token")
    _auth_token = request.GET.get("auth_token")

    try:
        if not (_request_token and _auth_token):
            raise Exception("parameter missing")

        # make sure the request token is generated from this server
        userinfo = RequestToken.load_info(_request_token)
        if not userinfo:
            raise Exception("request token not found")
        userobj = json.loads(userinfo)
        if not userobj:
            raise Exception("request token error")
        RequestToken.revoke(_request_token)  # delete used request token

        # verify auth token
        if _auth_token != gen_auth_token(_request_token):
            raise Exception("auth token error")

        return JsonResponse({"user": userobj})
    except Exception as e:
        # return JsonResponse({"error": str(e)}, status=403)
        return JsonResponse403({"error": str(e)})


def login(request):
    redirect_to = request.POST.get("next", request.GET.get("next"))
    request_token = request.POST.get("request_token", request.GET.get("request_token"))
    api_key = request.POST.get("api_key", request.GET.get("api_key"))

    if not (redirect_to and request_token and api_key):
        return HttpResponse("parameter missing", status=403)

    rt = RequestToken.load_info(request_token)
    if not rt:
        return HttpResponse("invalid request token", status=403)

    # verify remote host
    url_info_to = urlparse.urlparse(redirect_to)
    try:
        SSO.objects.get(host=url_info_to.netloc, credential__apikey=api_key)
    except SSO.DoesNotExist:
        return HttpResponse("request deny for remote host", status=403)

    okresp = HttpResponseRedirect(append_tokens(redirect_to,
                                                request_token=request_token,
                                                auth_token=gen_auth_token(request_token)))

    ssouser = request.user  # current logged-in user by default

    if request.method == "POST":
        form = AuthenticationForm(request, data=request.POST)
        if form.is_valid():
            RequestToken.bind_info(request_token, form.get_user())
            return okresp
    else:  # GET
        form = AuthenticationForm(request)
        bContinueWithCurrentAccount = request.GET.get("cwca", None)

        if bContinueWithCurrentAccount == "yes":  # confirmed: continue with logged in account
            RequestToken.bind_info(request_token, request.user)
            return okresp
        elif bContinueWithCurrentAccount == "no":  # confirmed: login with another account
            ssouser = AnonymousUser()
        else:  # ask to login with current account or switch to a new account, or login form if not logged-in
            pass

    current_site = get_current_site(request)

    context = {
        'form': form,
        "next": redirect_to,
        'site': current_site,
        'site_name': current_site.name,
        'remote_app': url_info_to.netloc,
        "sso_timeout": SSO_REQUEST_TOKEN_TTL,
        "ssouser": ssouser,
        "req": request  # name req will not confilict with request if django.core.context_processors.request enabled

    }

    return TemplateResponse(request, '/djssoserver/ssologin.html', context)



import hmac
import hashlib
import json
import uuid

from django.forms.models import model_to_dict
from django.conf import settings
from django.core.cache import cache
from . import SSO_SERVER_USER_MODEL_TO_DICT_CLS
from django.utils.module_loading import import_by_path


SSO_SEC_KEY = settings.SECRET_KEY
SSO_REQUEST_TOKEN_TTL = 240


def gen_auth_token(request_token):
    return hmac.new(str(SSO_SEC_KEY), request_token, digestmod=hashlib.sha256).hexdigest()


class RequestToken(object):
    @staticmethod
    def bind_info(request_token, user):
        # bind request token with user object
        cachedreqtoken = RequestToken.gen_cached_key(request_token)
        cache.set(cachedreqtoken, json.dumps(
            model_to_dict(user, exclude=["password", "user_permissions"]),
            cls=import_by_path(SSO_SERVER_USER_MODEL_TO_DICT_CLS)))

    @staticmethod
    def load_info(_request_token):
        return cache.get(RequestToken.gen_cached_key(_request_token))

    @staticmethod
    def generate():
        rt = uuid.uuid4().hex[:8]
        cache.set(RequestToken.gen_cached_key(rt), "{}", SSO_REQUEST_TOKEN_TTL)
        return rt

    @staticmethod
    def gen_cached_key(_request_token):
        return "sso::%s" % _request_token

    @staticmethod
    def revoke(_request_token):
        cache.delete(RequestToken.gen_cached_key(_request_token))



from django.db import models
from django.conf import settings
from django.utils.module_loading import import_by_path
from django.db.models import Q

from .utility import generalcache
from djapiauth.models import APIEntryPoint, APIKeys
from djapiauth.utility import traverse_urls
from . import MAGIC_ID


class SSO(models.Model):
    class Meta:
        verbose_name = "SSO credential"

    credential = models.ForeignKey(APIKeys)
    host = models.CharField(max_length=100, unique=True)
    note = models.CharField(max_length=80, null=True, blank=True)

    @staticmethod
    @generalcache(cachekey="ssoapi")
    def sso_api_list():
        """
        return sso related API
        """
        ssourls = []

        def collect(u, prefixre, prefixname):
            _prefixname = prefixname + [u._regex, ]
            urldisplayname = " ".join(_prefixname)

            if hasattr(u.urlconf_module, "_MODULE_MAGIC_ID_") \
                    and getattr(u.urlconf_module, "_MODULE_MAGIC_ID_") == MAGIC_ID:  # find aap name sso
                ssourls.append(urldisplayname)

        rooturl = import_by_path(settings.ROOT_URLCONF + ".urlpatterns")
        # traverse url matching tree to find the url statement including sso app,
        # should be 1 element in the ssourls unless you assigned 1+ prefix url for sso app
        traverse_urls(rooturl, resolverFunc=collect)

        finalQ = Q()  # filter to get full url of all registered sso api
        for prefix in ssourls:
            finalQ |= Q(name__startswith=prefix)
        return APIEntryPoint.objects.filter(finalQ)

    def save(self, *args, **kwargs):
        if not self.id:  # new creating
            c = APIKeys(note="for host: %s SSO login" % self.host)
            c.save()
            c.apis = SSO.sso_api_list()
            self.credential = c
        return super(SSO, self).save(*args, **kwargs)

    def __unicode__(self):
        return unicode(self.host)



from django.conf import settings
default_app_config = 'djssoserver.apps.SSOConfig'
MAGIC_ID = "sso19841108"  # used to detect app's position in an url tree


def _load_setting(n, default):
    return getattr(settings, n) if hasattr(settings, n) else default

SSO_SERVER_USER_MODEL_TO_DICT_CLS = _load_setting("SSO_SERVER_USER_MODEL_TO_DICT_CLS", "djssoserver.utility.DjangoJSONEncoder")


from django.contrib import admin
from django.forms import ModelForm
from .models import SSO


class SSOForm(ModelForm):
    class Meta:
        model = SSO
        exclude = ("credential",)


class SSOAdmin(admin.ModelAdmin):
    form = SSOForm
    list_display = ('host', 'note', 'get_apikey', 'get_seckey')

    def get_apikey(self, obj):
        return obj.credential.apikey

    def get_seckey(self, obj):
        return obj.credential.seckey

    get_apikey.short_description = 'API key'
    get_seckey.short_description = 'SEC key'


admin.site.register(SSO, SSOAdmin)



import cPickle
import json
import datetime
import decimal
import urllib
import urlparse

from django.core.cache import cache
from django.http import HttpResponse


def is_aware(value):
    """
    Determines if a given datetime.datetime is aware.

    The logic is described in Python's docs:
    http://docs.python.org/library/datetime.html#datetime.tzinfo
    """
    return value.tzinfo is not None and value.tzinfo.utcoffset(value) is not None


class DjangoJSONEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time and decimal types.
    """

    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime.datetime):
            r = o.isoformat()
            if o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith('+00:00'):
                r = r[:-6] + 'Z'
            return r
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif isinstance(o, datetime.time):
            if is_aware(o):
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r
        elif isinstance(o, decimal.Decimal):
            return str(o)
        else:
            return super(DjangoJSONEncoder, self).default(o)


class JsonResponse(HttpResponse):
    """
    An HTTP response class that consumes data to be serialized to JSON.

    :param data: Data to be dumped into json. By default only ``dict`` objects
      are allowed to be passed due to a security flaw before EcmaScript 5. See
      the ``safe`` parameter for more information.
    :param encoder: Should be an json encoder class. Defaults to
      ``django.core.serializers.json.DjangoJSONEncoder``.
    :param safe: Controls if only ``dict`` objects may be serialized. Defaults
      to ``True``.
    """

    def __init__(self, data, encoder=DjangoJSONEncoder, safe=True, **kwargs):
        if safe and not isinstance(data, dict):
            raise TypeError('In order to allow non-dict objects to be '
                            'serialized set the safe parameter to False')
        kwargs.setdefault('content_type', 'application/json')
        data = json.dumps(data, cls=encoder)
        super(JsonResponse, self).__init__(content=data, **kwargs)


def jsonify(function=None, options={}):
    def real_decorator(func):
        def wrapped(*args, **kwargs):
            resp = HttpResponse(json.dumps(func(*args, **kwargs), **options))
            resp["Content-Type"] = "application/json"
            return resp

        return wrapped

    return real_decorator if not function else real_decorator(function)


def generalcache(cachekey, ttl=1200):
    def real_decorator(func):
        def wrapper(*args, **kwargs):
            cachedvalue = cache.get(cachekey)
            if cachedvalue:
                return cPickle.loads(cachedvalue)
            else:
                result = func(*args, **kwargs)
                cache.set(cachekey, cPickle.dumps(result), ttl)
                return result
        return wrapper
    return real_decorator


def append_tokens(url, **kwargs):
    url_parts = list(urlparse.urlparse(url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update(kwargs)
    url_parts[4] = urllib.urlencode(query)
    return urlparse.urlunparse(url_parts)